==> flowtask-5.5.4-cp310-cp310-manylinux_2_17_x86_64.manylinux2014_x86_64.whl: flowtask/components/BestBuy.py <==
"""
Scrapping a Web Page Using Selenium + ChromeDriver + BeautifulSoup.
"""
import asyncio
from collections.abc import Callable
import random
from bs4 import BeautifulSoup
import httpx
import pandas as pd
import backoff
from navconfig.logging import logging
# Internals
from ..exceptions import (
    ComponentError,
    DataNotFound,
    NotSupported,
    ConfigError
)
from .abstract import FlowComponent
from ..interfaces import HTTPService, SeleniumService
from ..interfaces.http import ua


logging.getLogger(name='selenium.webdriver').setLevel(logging.WARNING)
logging.getLogger(name='WDM').setLevel(logging.WARNING)
logging.getLogger(name='hpack').setLevel(logging.WARNING)
logging.getLogger(name='seleniumwire').setLevel(logging.WARNING)


ProductPayload = {
    "locationId": None,
    "zipCode": None,
    "showOnShelf": True,
    "lookupInStoreQuantity": True,
    "xboxAllAccess": False,
    "consolidated": True,
    "showOnlyOnShelf": False,
    "showInStore": True,
    "pickupTypes": [
        "UPS_ACCESS_POINT",
        "FEDEX_HAL"
    ],
    "onlyBestBuyLocations": True,
    "items": [
        {
            "sku": None,
            "condition": None,
            "quantity": 1,
            "itemSeqNumber": "1",
            "reservationToken": None,
            "selectedServices": [],
            "requiredAccessories": [],
            "isTradeIn": False,
            "isLeased": False
        }
    ]
}

class BestBuy(FlowComponent, HTTPService, SeleniumService):
    """BestBuy.

    Combining API Key and Web Scrapping, this component will be able to extract
    Best Buy Information (stores, products, Product Availability, etc).
    """
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop = None,
        job: Callable = None,
        stat: Callable = None,
        **kwargs,
    ):
        self._fn = kwargs.pop('type', None)
        self.chunk_size: int = kwargs.get('chunk_size', 100)
        self.product_info: bool = kwargs.get('product_info', False)
        if not self._fn:
            raise ConfigError(
                "BestBuy: require a `type` Function to be called, ex: availability"
            )
        super(BestBuy, self).__init__(
            loop=loop,
            job=job,
            stat=stat,
            **kwargs
        )
        # Always use proxies:
        self.use_proxy: bool = True
        self._free_proxy: bool = False
        ctt_list: list = [
            "f3dbf688e45146555bb2b8604a993601",
            "06f4dfe367e87866397ef32302f5042e",
            "4e07e03ff03f5debc4e09ac4db9239ac"
        ]
        sid_list: list = [
            "d4fa1142-2998-4b68-af78-46d821bb3e1f",
            "9627390e-b423-459f-83ee-7964dd05c9a8"
        ]
        self.cookies = {
            # "CTT": ,
            "CTT": random.choice(ctt_list),
            "SID": random.choice(sid_list),
            "bby_rdp": "l",
            "bm_sz": "9F5ED0110AF18594E2347A89BB4AB998~YAAQxm1lX6EqYHGSAQAAw+apmhkhXIeGYEc4KnzUMsjeac3xEoQmTNz5+of62i3RXQL6fUI+0FvCb/jgSjiVQOcfaSF+LdLkOXP1F4urgeIcqp/dBAhu5MvZXaCQsT06bwr7j21ozhFfTTWhjz1HmZN8wecsE6WGbK6wXp/33ODKlLaGWkTutqHbkzvMiiHXBCs9hT8jVny0REfita4AfqTK85Y6/M6Uq4IaDLPBLnTtJ0cTlPHk1HmkG5EsnI46llghcx1KZnCGnvZfHdb2ME9YZJ2GmC2b7dNmAgyL/gSVpoNdCJOj5Jk6z/MCVhZ81OZfX4S01E2F1mBGq4uV5/1oK2KR4YgZP4dsTN8izEEPybUKGY3CyM1gOUc=~3556420~4277810",  # noqa
            "bby_cbc_lb": "p-browse-e",
            "intl_splash": "false"
        }
        self.headers: dict = {
            "Host": "www.bestbuy.com",
            "Referer": "https://www.bestbuy.com/",
            "X-Requested-With": "XMLHttpRequest",
            "TE": "trailers",
            "Accept-Language": "en-US,en;q=0.5",
            # "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",  # noqa
            "Accept-Encoding": "gzip, deflate",
            "DNT": "1",
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
            "User-Agent": random.choice(ua),
            **self.headers
        }
        self.semaphore = asyncio.Semaphore(10)

    async def close(self, **kwargs) -> bool:
        self.close_driver()
        return True

    async def start(self, **kwargs) -> bool:
        await super(BestBuy, self).start(**kwargs)
        if self.previous:
            self.data = self.input
        else:
            raise DataNotFound(
                "Data Not Found",
                code=404
            )
        if not isinstance(self.data, pd.DataFrame):
            raise ComponentError(
                "Incompatible Pandas Dataframe",
                code=404
            )
        if self._fn == 'availability':
            if not hasattr(self, 'brand'):
                raise ConfigError(
                    "BestBuy: A Brand is required for using Product Availability"
                )
            # if not hasattr(self, 'sku'):
            #     raise ConfigError(
            #         "BestBuy: Product SKU is required for using Product Availability"
            #     )
        if not hasattr(self, self._fn):
            raise ConfigError(
                f"BestBuy: Unable to found Function {self._fn} in BBY Component."
            )

    def _search_product(self, brand: str, sku: str) -> str:
        front_url = "https://www.bestbuy.com/site/searchpage.jsp?cp="
        middle_url = "&searchType=search&st="
        page_count = 1
        # TODO: Get the Brand and Model from the Component.
        search_term = f'{brand}%20{sku}'
        end_url = "&_dyncharset=UTF-8&id=pcat17071&type=page&sc=Global&nrp=&sp=&qp=&list=n&af=true&iht=y&usc=All%20Categories&ks=960&keys=keys"  # noqa
        url = front_url + str(page_count) + middle_url + search_term + end_url
        print('SEARCH URL: ', url)
        return url

    async def _extract_product_list(self, content: str) -> dict:
        soup = BeautifulSoup(content, 'html.parser')
        # Find all elements with class "sku-item"
        product_items = soup.find_all('li', {'class': ['sku-item']})
        # Iterate over each product item
        result = {}
        for item in product_items:
            # Get the "data-sku-id" attribute
            sku_id = item.get("data-sku-id")
            # Check if the SKU ID matches your target SKU ID
            if sku_id == self.sku:
                print(f"Found matching SKU ID: {sku_id}")
                # Now look for the child with class "sku-title"
                pd = item.find('h4', {'class': ['sku-title']})
                anchor = pd.a
                title = anchor.text
                price = item.find(
                    'div', {'class': ['priceView-customer-price']}
                ).select_one(
                    'span:nth-child(1)'
                ).text
                # Image:
                image = item.find('img', {"class": ['product-image']})['src']
                # Product Model
                model = item.select_one(
                    'div.sku-attribute-title:nth-child(1) span.sku-value'
                ).text
                url = "https://www.bestbuy.com{url}".format(
                    url=anchor['href']
                )
                self._logger.notice(
                    f':: Product URL: {url}'
                )
                result = {
                    "sku": sku_id,
                    "brand": self.brand,
                    "model": model,
                    "product_name": title,
                    "image_url": image,
                    "price": price,
                    "url": url
                }
                # if product found, break
                return result
        return result

    def chunkify(self, lst, n):
        """Split list lst into chunks of size n."""
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    async def api_post(self, url: str, cookies: httpx.Cookies, payload: dict):
        proxies = await self.get_proxies()
        proxy = proxies[0]
        proxies = {
            "http://": httpx.AsyncHTTPTransport(
                proxy=f"http://{proxy}"
            ),
            "https://": httpx.AsyncHTTPTransport(
                proxy=f"http://{proxy}"
            ),
        }
        timeout = httpx.Timeout(connect=5.0, read=10.0, write=5.0, pool=20.0)
        async with httpx.AsyncClient(cookies=cookies, verify=False) as client:
            try:
                response = await client.post(
                    url,
                    json=payload,
                    headers=self.headers,
                    timeout=timeout
                )
                response.raise_for_status()
                if response.status_code == 200:
                    return response.json()
                return {}
            except httpx.TimeoutException:
                raise
            except (httpx.HTTPError) as ex:
                print('ERR > ', ex)
                raise
            except Exception as exc:
                print('EXC > ', exc)
                raise ComponentError(
                    f"An error occurred: {exc}"
                )

    # @backoff.on_exception(
    #     backoff.expo,
    #     (httpx.ConnectTimeout, asyncio.TimeoutError),
    #     max_tries=2
    # )
    async def _check_store_availability(
        self,
        sku,
        row,
        cookies
    ):
        async with self.semaphore:
            # extract zipcode, sku and location_code from row:
            zipcode = row['zipcode']
            location_code = row['location_code']
            payload = ProductPayload.copy()
            payload["locationId"] = location_code
            payload["zipCode"] = zipcode
            # Useful for looking several products at time
            for item in payload["items"]:
                item["sku"] = sku
            try:
                result = await self.api_post(
                    url="https://www.bestbuy.com/productfulfillment/c/api/2.0/storeAvailability",
                    cookies=cookies,
                    payload=payload
                )
            except httpx.TimeoutException as ex:
                self._logger.error(f"Request timed out: {ex}")
                return row
            except httpx.HTTPError as ex:
                self._logger.error(ex)
                return row
            except Exception as ex:
                self._logger.error(ex)
                return row
            if not result:
                return row
            # working with result:
            # TODO: can work with several products at the same time:
            items = result.get('ispu', {}).get('items', [])
            # for every item, extract the SKU and locations:
            for item in items:
                # product_sku = item.get('sku')
                locations = item.get('locations')
                for location in locations:
                    lid = location.get('locationId')
                    matching_store = self.data[self.data['location_code'] == lid]
                    # If we found a matching store (location_code == locationId)
                    if not matching_store.empty:
                        # Get the index of the matched row
                        idx = matching_store.index[0]
                        # adding Brand:
                        self.data.at[idx, 'brand'] = self.brand
                        # and location data:
                        self.data.at[idx, 'location_data'] = location
                        # Update the row with the location data
                        # by converting location dict to columns
                        for key, val in location.items():
                            self.data.at[idx, key] = val
                            if key == 'inStoreAvailability':
                                try:
                                    self.data.at[idx, 'availableInStoreQuantity'] = val.get('availableInStoreQuantity')
                                except KeyError:
                                    pass
                        # Mark the row as checked
                        self.data.at[idx, 'checked'] = True
                        print(
                            f"Updated Store with LocationId {lid} at index {idx}"
                        )
            # To avoid being blocked, sleep for a while, random time between 1 and 3 seconds
            await asyncio.sleep(
                random.randint(1, 3)
            )
            # Return the Row
            return row

    def column_exists(self, column: str):
        if column not in self.data.columns:
            self._logger.warning(
                f"Column {column} does not exist in the dataframe"
            )
            self.data[column] = None
            return False
        return True

    async def availability(self):
        """availability.

        Best Buy Product Availability.
        """
        httpx_cookies = httpx.Cookies()
        if self.product_info is True:
            # 1. Get the Product List URL
            url = self._search_product(self.brand, self.sku)
            await self.get_page(url, self.cookies)
            content = self.driver().page_source
            product_info = await self._extract_product_list(content)
            if not product_info:
                raise DataNotFound(
                    f"BestBuy: Cannot found product information for {self.brand}-{self.sku}"
                )
            # With Product info, extract the cookies from Selenium:
            # Get the cookies from Selenium WebDriver
            selenium_cookies = self.driver().get_cookies()
            # Convert Selenium cookies to httpx format
            cookies_avail = ["CTT", "SID", '_abck', 'bby_rdp', 'bm_sz']
            for cookie in selenium_cookies:
                if cookie['name'] in cookies_avail:
                    httpx_cookies.set(
                        cookie['name'], cookie['value'],
                        domain=cookie['domain'],
                        path=cookie['path']
                    )

            # make a join between current df and product info:
            new_df = pd.DataFrame([product_info] * len(self.data))
            # Concatenate the DataFrames horizontally
            self.data = pd.concat([self.data, new_df], axis=1)
        else:
            for key, value in self.cookies.items():
                httpx_cookies.set(
                    key, value,
                    domain='.bestbuy.com',
                    path='/'
                )

        # With available cookies, iterate over dataframe for stores:
        self.data['checked'] = False  # Add 'checked' flag column

        # define the columns returned:
        self.column_exists('brand')
        self.column_exists('location_data')
        self.column_exists('locationId')
        self.column_exists('availability')
        self.column_exists('inStoreAvailability')
        self.column_exists('onShelfDisplay')
        self.column_exists('availableInStoreQuantity')

        for _, row in self.data.iterrows():
            if not row['checked']:
                if not hasattr(self, 'sku'):
                    self.sku = row['sku']
                await self._check_store_availability(
                    self.sku, row, httpx_cookies
                )

        # Remove the 'checked' column after completion
        # self.data.drop(columns=['checked'], inplace=True)
        # return existing data
        return self.data

    async def run(self):
        # we need to call the "function" for Best Buy Services.
        fn = getattr(self, self._fn)
        result = None
        try:
            result = await fn()
        except (ComponentError, TimeoutError, NotSupported):
            raise
        except Exception as exc:
            raise ComponentError(
                f"BestBuy: Unknown Error: {exc}"
            ) from exc
        # Print results
        print(result)
        print("::: Printing Column Information === ")
        for column, t in result.dtypes.items():
            print(column, "->", t, "->", result[column].iloc[0])
        self._result = result
        return self._result

==> flowtask-5.5.10-cp310-cp310-manylinux_2_17_x86_64.manylinux2014_x86_64.whl: flowtask/components/BestBuy.py <==
"""
Scrapping a Web Page Using Selenium + ChromeDriver + BeautifulSoup.
"""
import asyncio
from collections.abc import Callable
import random
from bs4 import BeautifulSoup
import httpx
import pandas as pd
import backoff
import ssl
from navconfig.logging import logging
# Internals
from ..exceptions import (
    ComponentError,
    DataNotFound,
    NotSupported,
    ConfigError
)
from .flow import FlowComponent
from ..interfaces import HTTPService, SeleniumService
from ..interfaces.http import ua


logging.getLogger(name='selenium.webdriver').setLevel(logging.WARNING)
logging.getLogger(name='WDM').setLevel(logging.WARNING)
logging.getLogger(name='hpack').setLevel(logging.WARNING)
logging.getLogger(name='seleniumwire').setLevel(logging.WARNING)


ProductPayload = {
    "locationId": None,
    "zipCode": None,
    "showOnShelf": True,
    "lookupInStoreQuantity": True,
    "xboxAllAccess": False,
    "consolidated": True,
    "showOnlyOnShelf": False,
    "showInStore": True,
    "pickupTypes": [
        "UPS_ACCESS_POINT",
        "FEDEX_HAL"
    ],
    "onlyBestBuyLocations": True,
    "items": [
        {
            "sku": None,
            "condition": None,
            "quantity": 1,
            "itemSeqNumber": "1",
            "reservationToken": None,
            "selectedServices": [],
            "requiredAccessories": [],
            "isTradeIn": False,
            "isLeased": False
        }
    ]
}

class BestBuy(FlowComponent, SeleniumService, HTTPService):
    """BestBuy.

    Combining API Key and Web Scrapping, this component will be able to extract
    Best Buy Information (stores, products, Product Availability, etc).
    """
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop = None,
        job: Callable = None,
        stat: Callable = None,
        **kwargs,
    ):
        self._fn = kwargs.pop('type', None)
        self.chunk_size: int = kwargs.get('chunk_size', 100)
        self.product_info: bool = kwargs.get('product_info', False)
        if not self._fn:
            raise ConfigError(
                "BestBuy: require a `type` Function to be called, ex: availability"
            )
        super(BestBuy, self).__init__(
            loop=loop,
            job=job,
            stat=stat,
            **kwargs
        )
        # Always use proxies:
        self.use_proxy: bool = True
        self._free_proxy: bool = False
        ctt_list: list = [
            "f3dbf688e45146555bb2b8604a993601",
            "06f4dfe367e87866397ef32302f5042e",
            "4e07e03ff03f5debc4e09ac4db9239ac"
        ]
        sid_list: list = [
            "d4fa1142-2998-4b68-af78-46d821bb3e1f",
            "9627390e-b423-459f-83ee-7964dd05c9a8"
        ]
        self.cookies = {
            # "CTT": ,
            "CTT": random.choice(ctt_list),
            "SID": random.choice(sid_list),
            "bby_rdp": "l",
            "bm_sz": "9F5ED0110AF18594E2347A89BB4AB998~YAAQxm1lX6EqYHGSAQAAw+apmhkhXIeGYEc4KnzUMsjeac3xEoQmTNz5+of62i3RXQL6fUI+0FvCb/jgSjiVQOcfaSF+LdLkOXP1F4urgeIcqp/dBAhu5MvZXaCQsT06bwr7j21ozhFfTTWhjz1HmZN8wecsE6WGbK6wXp/33ODKlLaGWkTutqHbkzvMiiHXBCs9hT8jVny0REfita4AfqTK85Y6/M6Uq4IaDLPBLnTtJ0cTlPHk1HmkG5EsnI46llghcx1KZnCGnvZfHdb2ME9YZJ2GmC2b7dNmAgyL/gSVpoNdCJOj5Jk6z/MCVhZ81OZfX4S01E2F1mBGq4uV5/1oK2KR4YgZP4dsTN8izEEPybUKGY3CyM1gOUc=~3556420~4277810",  # noqa
            "bby_cbc_lb": "p-browse-e",
            "intl_splash": "false"
        }
        self.headers: dict = {
            "Host": "www.bestbuy.com",
            "Referer": "https://www.bestbuy.com/",
            "X-Requested-With": "XMLHttpRequest",
            "TE": "trailers",
            "Accept-Language": "en-US,en;q=0.5",
            # "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",  # noqa
            "Accept-Encoding": "gzip, deflate",
            "DNT": "1",
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
            "User-Agent": random.choice(ua),
            **self.headers
        }
        self.semaphore = asyncio.Semaphore(10)

    async def close(self, **kwargs) -> bool:
        self.close_driver()
        return True

    async def start(self, **kwargs) -> bool:
        await super(BestBuy, self).start(**kwargs)
        if self.previous:
            self.data = self.input
        else:
            raise DataNotFound(
                "Data Not Found",
                code=404
            )
        if not isinstance(self.data, pd.DataFrame):
            raise ComponentError(
                "Incompatible Pandas Dataframe",
                code=404
            )
        if self._fn == 'availability':
            if not hasattr(self, 'brand'):
                raise ConfigError(
                    "BestBuy: A Brand is required for using Product Availability"
                )
        if not hasattr(self, self._fn):
            raise ConfigError(
                f"BestBuy: Unable to found Function {self._fn} in BBY Component."
            )

    def _search_product(self, brand: str, sku: str) -> str:
        front_url = "https://www.bestbuy.com/site/searchpage.jsp?cp="
        middle_url = "&searchType=search&st="
        page_count = 1
        # TODO: Get the Brand and Model from the Component.
        search_term = f'{brand}%20{sku}'
        end_url = "&_dyncharset=UTF-8&id=pcat17071&type=page&sc=Global&nrp=&sp=&qp=&list=n&af=true&iht=y&usc=All%20Categories&ks=960&keys=keys"  # noqa
        url = front_url + str(page_count) + middle_url + search_term + end_url
        print('SEARCH URL: ', url)
        return url

    async def _extract_product_list(self, content: str) -> dict:
        soup = BeautifulSoup(content, 'html.parser')
        # Find all elements with class "sku-item"
        product_items = soup.find_all('li', {'class': ['sku-item']})
        # Iterate over each product item
        result = {}
        for item in product_items:
            # Get the "data-sku-id" attribute
            sku_id = item.get("data-sku-id")
            # Check if the SKU ID matches your target SKU ID
            if sku_id == self.sku:
                print(f"Found matching SKU ID: {sku_id}")
                # Now look for the child with class "sku-title"
                pd = item.find('h4', {'class': ['sku-title']})
                anchor = pd.a
                title = anchor.text
                price = item.find(
                    'div', {'class': ['priceView-customer-price']}
                ).select_one(
                    'span:nth-child(1)'
                ).text
                # Image:
                image = item.find('img', {"class": ['product-image']})['src']
                # Product Model
                model = item.select_one(
                    'div.sku-attribute-title:nth-child(1) span.sku-value'
                ).text
                url = "https://www.bestbuy.com{url}".format(
                    url=anchor['href']
                )
                self._logger.notice(
                    f':: Product URL: {url}'
                )
                result = {
                    "sku": sku_id,
                    "brand": self.brand,
                    "model": model,
                    "product_name": title,
                    "image_url": image,
                    "price": price,
                    "url": url
                }
                # if product found, break
                return result
        return result

    def chunkify(self, lst, n):
        """Split list lst into chunks of size n."""
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    async def api_post(self, url: str, cookies: httpx.Cookies, payload: dict):
        self._free_proxy = False
        proxies = await self.get_proxies()
        proxy = proxies[0]
        # proxies = {
        #     "http://": httpx.AsyncHTTPTransport(
        #         proxy=f"http://{proxy}"
        #     ),
        #     "https://": httpx.AsyncHTTPTransport(
        #         proxy=f"http://{proxy}"
        #     ),
        # }
        proxy_config = {
            "http://": f"http://{proxy}",
            "https://": f"http://{proxy}"  # Using the same proxy for HTTPS
        }

        # Define custom SSL context
        ssl_context = ssl.create_default_context()
        # Disable older protocols if needed
        ssl_context.options |= ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
        # Ensure at least TLS 1.2 is used
        ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

        # Use AsyncHTTPTransport to pass in SSL context explicitly
        transport = httpx.AsyncHTTPTransport(retries=2, verify=ssl_context)
        timeout = httpx.Timeout(connect=5.0, read=20.0, write=5.0, pool=20.0)

        async with httpx.AsyncClient(
            cookies=cookies,
            proxies=proxy_config,
            transport=transport
        ) as client:
            try:
                response = await client.post(
                    url,
                    json=payload,
                    headers=self.headers,
                    timeout=timeout
                )
                response.raise_for_status()
                if response.status_code == 200:
                    return response.json()
                else:
                    print(
                        f"API request failed with status code {response.status_code}"
                    )
                return {}
            except httpx.TimeoutException:
                raise
            except (httpx.HTTPError) as ex:
                raise httpx.HTTPError(ex)
            except Exception as exc:
                print('EXC > ', exc)
                raise ComponentError(
                    f"An error occurred: {exc}"
                )

    # @backoff.on_exception(
    #     backoff.expo,
    #     (httpx.ConnectTimeout, httpx.ReadTimeout),
    #     max_tries=2
    # )
    async def _check_store_availability(self, row, cookies):
        async with self.semaphore:
            # Prepare payload for the API request
            zipcode = row['zipcode']
            location_code = row['location_code']
            sku = row['sku']
            payload = ProductPayload.copy()
            payload["locationId"] = location_code
            payload["zipCode"] = zipcode
            for item in payload["items"]:
                item["sku"] = sku
            try:
                result = await self.api_post(
                    url="https://www.bestbuy.com/productfulfillment/c/api/2.0/storeAvailability",
                    cookies=cookies,
                    payload=payload
                )
            except (httpx.TimeoutException, httpx.HTTPError) as ex:
                self._logger.warning(f"Request failed: {ex}")
                return row
            except Exception as ex:
                self._logger.error(f"An error occurred: {ex}")
                return row

            if not result:
                self._logger.warning(
                    f"No availability data found for {sku} at zipcode {zipcode}"
                )
                return row

            # Extract the availability data from the API response
            items = result.get('ispu', {}).get('items', [])
            for item in items:
                locations = item.get('locations', [])
                for location in locations:
                    lid = location.get('locationId')
                    # Find matching store and SKU in DataFrame
                    matching_store = self.data[
                        (self.data['location_code'] == lid) & (self.data['sku'] == sku)
                    ]
                    if not matching_store.empty:
                        idx = matching_store.index[0]
                        if self.data.loc[idx, 'checked'] == True:
                            print('Already checked, continue ...')
                            continue  # Skip this row if it's already marked as checked
                        print(f'Found matching store {lid} for sku {sku}')
                        # Update the DataFrame row with new availability data
                        self.data.loc[idx, ['brand', 'location_data']] = [self.brand, location]
                        for key, val in location.items():
                            if key in self.data.columns:
                                self.data.at[idx, key] = val
                            else:
                                self.data.at[idx, key] = None
                            if key == 'inStoreAvailability':
                                try:
                                    self.data.loc[idx, 'availableInStoreQuantity'] = val.get(
                                        'availableInStoreQuantity', 0
                                    )
                                except KeyError:
                                    self.data.loc[idx, 'availableInStoreQuantity'] = None
                        # Mark the row as checked
                        self.data.loc[idx, 'checked'] = True
            return row

    def column_exists(self, column: str):
        if column not in self.data.columns:
            self._logger.warning(
                f"Column {column} does not exist in the Dataframe"
            )
            self.data[column] = None
            return False
        return True

    async def availability(self):
        """availability.

        Best Buy Product Availability.
        """
        httpx_cookies = httpx.Cookies()
        if self.product_info is True:
            # 1. Get the Product List URL
            url = self._search_product(self.brand, self.sku)
            await self.get_page(url, self.cookies)
            content = self.driver().page_source
            product_info = await self._extract_product_list(content)
            if not product_info:
                raise DataNotFound(
                    f"BestBuy: Cannot found product information for {self.brand}-{self.sku}"
                )
            # With Product info, extract the cookies from Selenium:
            # Get the cookies from Selenium WebDriver
            selenium_cookies = self.driver().get_cookies()
            # Convert Selenium cookies to httpx format
            cookies_avail = ["CTT", "SID", '_abck', 'bby_rdp', 'bm_sz']
            for cookie in selenium_cookies:
                if cookie['name'] in cookies_avail:
                    httpx_cookies.set(
                        cookie['name'], cookie['value'],
                        domain=cookie['domain'],
                        path=cookie['path']
                    )

            # make a join between current df and product info:
            new_df = pd.DataFrame([product_info] * len(self.data))
            # Concatenate the DataFrames horizontally
            self.data = pd.concat([self.data, new_df], axis=1)
        else:
            for key, value in self.cookies.items():
                httpx_cookies.set(
                    key, value,
                    domain='.bestbuy.com',
                    path='/'
                )

        # With available cookies, iterate over dataframe for stores:
        self.data['checked'] = False  # Add 'checked' flag column

        # define the columns returned:
        self.column_exists('brand')
        self.column_exists('location_data')
        self.column_exists('locationId')
        self.column_exists('availability')
        self.column_exists('inStoreAvailability')
        self.column_exists('onShelfDisplay')
        self.column_exists('availableInStoreQuantity')

        # Iterate over each row in the DataFrame
        print('starting ...')
        # Fetch the rows where 'checked' is False
        unchecked_rows = self.data.loc[self.data['checked'] == False]

        # Break the loop if no unchecked rows are left
        if unchecked_rows.empty:
            return self.data

        # Iterate over unchecked rows
        for idx, row in unchecked_rows.iterrows():
            if self.data.loc[idx, 'checked'] == True:
                continue  # Skip this row if it's already marked as checked
            await self._check_store_availability(
                row, httpx_cookies
            )
            # if self.data.loc[idx, 'checked'] == True:
            #     print(f"Row at index {idx} marked as checked.")
            # else:
            #     print(f"Row at index {idx} remains unchecked.")
            await asyncio.sleep(
                random.uniform(0.5, 1.5)
            )

        # show the num of rows in final dataframe:
        self._logger.notice(
            "Ending Checking Availability."
        )
        # Remove the 'checked' column after completion
        # self.data.drop(columns=['checked'], inplace=True)
        # return existing data
        return self.data

    async def run(self):
        # we need to call the "function" for Best Buy Services.
        fn = getattr(self, self._fn)
        result = None
        if not callable(fn):
            raise ComponentError(
                f"Best Buy: Functino {self._fn} doesn't exists."
            )
        try:
            result = await fn()
        except (ComponentError, TimeoutError, NotSupported):
            raise
        except Exception as exc:
            raise ComponentError(
                f"BestBuy: Unknown Error: {exc}"
            ) from exc
        # Print results
        print(result)
        print("::: Printing Column Information === ")
        for column, t in result.dtypes.items():
            print(column, "->", t, "->", result[column].iloc[0])
        self._result = result
        return self._result
